import copy
import numpy as np

def addChildrenToList(joint, jointsToDelete):
    jointsToDelete.append(joint.name)
    for child in joint.children: 
        addChildrenToList(child, jointsToDelete)

def removeLimb(bvhData, jointName):
    bvhDataCopy = copy.deepcopy(bvhData)
    if(jointName == bvhDataCopy.skeleton.root.name):
        print(f"\033[1;33mWARNING\033[0m: you are trying to remove the root joint. You can't do this as this would return an empty BVH. Returning bvh unchanged.")
        return bvhDataCopy

    topJoint = bvhDataCopy.skeleton.getJoint(jointName)
    # create the list with names of joints to delete
    jointsToDelete = []
    jointsToDelete.append(jointName)
    for child in topJoint.children:
        addChildrenToList(child, jointsToDelete)

    motionColumnsToDelete = []
    # create the list with motion column numbers to delete
    for jointToDeleteName in jointsToDelete:
        joint = bvhDataCopy.skeleton.getJoint(jointToDeleteName)
        offset = 0
        for channel in joint.channels:
            motionColumnsToDelete.append(joint.motionIndex + offset)
            offset += 1

    # iterate over the names in reverse order
    for jointToDeleteName in reversed(jointsToDelete):
        joint = bvhDataCopy.skeleton.getJoint(jointToDeleteName)
        # REMOVE the JOINT FROM it's PARENTS list
        joint.parent.children = [child for child in joint.parent.children if child.name != jointToDeleteName]
        # REMOVE the JOINT itself
        del bvhDataCopy.skeleton.joints[jointToDeleteName]
    
    # REMOVE the necessary part of the MOTION columns
    bvhDataCopy.motion.frames = [[num for i, num in enumerate(frame) if i not in motionColumnsToDelete] for frame in bvhDataCopy.motion.frames]

    # REFRESH the indexes and motionIndexes for all joints and their respective dictionaries
    newSkeleton = bvhDataCopy.skeleton
    newSkeleton.jointIndexes = newSkeleton.buildJointIndexDict(newSkeleton.root, [0])
    newSkeleton.hierarchyIndexes = newSkeleton.buildHierarchyIndexDict(newSkeleton.root, [0])
    return bvhDataCopy
    
def scaleSkeleton(bvhData, scaleFactor):
    bvhDataCopy = copy.deepcopy(bvhData)
    if(scaleFactor<=0.0):
        print(f"\033[1;33mWARNING\033[0m: The scale factor has to be greater than 0. Returning bvh unchanged.")
        return bvhDataCopy
    for bone in bvhDataCopy.skeleton.joints.values():
        bone.offset = np.multiply(bone.offset, scaleFactor)
    rootJoint = bvhDataCopy.skeleton.root
    if any(rootJoint.getChannelIndex(axis) == 0 for axis in ["Xposition", "Yposition", "Zposition"]):
        positionSlice = slice(0,3)
    else:
        positionSlice = slice(3,6)
    for frame in bvhDataCopy.motion.frames:
        frame[positionSlice] = np.multiply(frame[positionSlice], scaleFactor)
    return bvhDataCopy


import argparse as ap
import sys
from bvhTools import bvhIO, bvhManipulation, bvhSlicer, bvhVisualizerMpl

def main():
    parser = ap.ArgumentParser(description="bvhTools CLI interface")
    subparsers = parser.add_subparsers(dest="command", help="Available commands", required=True)
    
    # subparser for centering commands
    parser_centering = subparsers.add_parser("center", help="Centering operations", description="Centers a BVH file's feet, root or the X and Z axes and writes it to a new BVH file")
    parser_centering.add_argument("--centeringOption", type=str, help="Option to execute", choices=["feet", "root", "xz"], default="feet")
    parser_centering.add_argument("--bvhFile", required=True, type=str, help="Path to the BVH file")
    parser_centering.add_argument("--outputFile", required=True, type=str, help="Path to the output BVH file")

    # subparser for rotation commands
    parser_rotation = subparsers.add_parser("rotate", help="Rotation commands", description="Rotates a BVH file in local or world space and writes it to a new BVH file")
    parser_rotation.add_argument("--rotationOption", type=str, help="Option to execute", choices=["local", "world"], default="local")
    parser_rotation.add_argument("--bvhFile", required=True, type=str, help="Path to the BVH file")
    parser_rotation.add_argument("--outputFile", required=True, type=str, help="Path to the output BVH file")
    parser_rotation.add_argument("--angles", required=True, type=float, nargs=3, help="Angles to rotate: X_value Y_value Z_value")

    # subparser for viewer commands
    parser_viewer = subparsers.add_parser("view", help="Viewer command", description="Launches matplotlib viewer. By default, it will show points, lines and labels. Disable them with --noPoints, --noLines and --noLabels")
    parser_viewer.add_argument("--bvhFile", required=True, type=str, help="Path to the BVH file")
    parser_viewer.add_argument("--noPoints", action="store_false", help="Do not draw points for joints")
    parser_viewer.add_argument("--noLines", action="store_false", help="Do not draw lines between joints")
    parser_viewer.add_argument("--noLabels", action="store_false", help="Do not draw labels")
    
    # subparser for slicing commands
    parser_slicer = subparsers.add_parser("slice", help="Slicer commands", description="Slices a BVH between two frames and writes it to a new BVH file")
    parser_slicer.add_argument("--bvhFile", required=True, type=str, help="Path to the BVH file")
    parser_slicer.add_argument("--startFrame", required=True, type=int, help="Start frame")
    parser_slicer.add_argument("--endFrame", required=True, type=int, help="End frame")
    parser_slicer.add_argument("--outputFile", required=True, type=str, help="Path to the output BVH file")

    # subparser for CSV commands
    parser_csv = subparsers.add_parser("csv", help="CSV commands", description="Exports a BVH file completely or its positions to a CSV file")
    parser_csv.add_argument("--csvOption", type=str, help="CSV export options: 'positions' for joint coordinates or 'bvh' for the whole BVH", choices=["positions", "bvh"], default="positions")
    parser_csv.add_argument("--bvhFile", required=True, type=str, help="Path to the BVH file")
    parser_csv.add_argument("--outputFile", required=True, type=str, help="Path to the output CSV file")

    args = parser.parse_args()

    commands = {
        "center": center_bvh,
        "rotate": rotate_bvh,
        "view": view_bvh,
        "slice": slice_bvh,
        "csv": csv_bvh
    }
    
    # dispatch the command
    if args.command in commands:
        commands[args.command](args)
    else:
        showHelp()
        sys.exit(1)

def showHelp():
    print("bvhTools: Command-line tools for BVH file manipulation\n")
    print("Usage:")
    print(f"  python {sys.argv[0]} <command> [options]\n")
    print("Commands:")
    print("  center     Center the skeleton (feet, root, or XZ)")
    print("  rotate     Apply rotation (local/world space)")
    print("  view       Launch a matplotlib-based BVH viewer")
    print("  slice      Extract a segment of the animation")
    print("  csv        Export joint position data or full BVH to CSV")
    print("\nUse '<command> --help' to see options for a specific command.")


def center_bvh(args):
    bvhData = bvhIO.readBvh(args.bvhFile)
    if(args.centeringOption == "feet"):
        bvhData = bvhManipulation.centerSkeletonFeet(bvhData)
    if(args.centeringOption == "root"):
        bvhData = bvhManipulation.centerSkeletonRoot(bvhData)
    if(args.centeringOption == "xz"):
        bvhData = bvhManipulation.centerSkeletonXZ(bvhData)
    bvhIO.writeBvh(bvhData, args.outputFile)

def rotate_bvh(args):
    bvhData = bvhIO.readBvh(args.bvhFile)
    if(args.rotationOption == "local"):
        bvhData = bvhManipulation.rotateSkeletonLocal(bvhData, args.angles)
    if(args.rotationOption == "world"):
        bvhData = bvhManipulation.rotateSkeletonWorld(bvhData, args.angles)
    bvhIO.writeBvh(bvhData, args.outputFile)

def view_bvh(args):
    bvhData = bvhIO.readBvh(args.bvhFile)
    bvhVisualizerMpl.showBvhAnimation(bvhData, showPoints=args.noPoints, showLines=args.noLines, showLabels=args.noLabels)

def slice_bvh(args):
    bvhData = bvhIO.readBvh(args.bvhFile)
    bvhData = bvhSlicer.getBvhSlice(bvhData, args.startFrame, args.endFrame)
    bvhIO.writeBvh(bvhData, args.outputFile)

def csv_bvh(args):
    bvhData = bvhIO.readBvh(args.bvhFile)
    if(args.csvOption == "positions"):
        bvhIO.writePositionsToCsv(bvhData, args.outputFile)
    if(args.csvOption == "bvh"):
        bvhIO.writeBvhToCsv(bvhData, args.outputFile)

if __name__ == "__main__":
    main()


from scipy.spatial.transform import Rotation as R
import numpy as np
import copy
import re

class Joint:
    def __init__(self, name, index, offset, channels, parent=None):
        self.name = name
        self.index = index
        self.motionIndex = -1
        self.offset = offset 
        self.channels = channels
        self.children = []
        self.parent = parent

    def setOffset(self, offset):
        self.offset = offset

    def setChannels(self, channels):
        self.channels = channels

    def setParent(self, parent):
        self.parent = parent

    def addChild(self, child):
        self.children.append(child)

    def getChannelCount(self):
        return len(self.channels)
    
    def getPositionChannelsOrder(self):
        if("position" not in self.channels[0] and len(self.channels) <= 3):
            print(f"\033[1;33mWARNING\033[0m: joint {self.name} has no position channels")
            return ""
        positionChannels = self.channels[0:3] if("position" in self.channels[0] or "position" in self.channels[1] or "position" in self.channels[2]) else self.channels[3:6]
        if(positionChannels[0] == "Xposition"):
            if(positionChannels[1] == "Yposition"):
                return "XYZ"
            if(positionChannels[1] == "Zposition"):
                return "XZY"
        if(positionChannels[0] == "Yposition"):
            if(positionChannels[1] == "Xposition"):
                return "YXZ"
            if(positionChannels[1] == "Zposition"):
                return "YZX"
        if(positionChannels[0] == "Zposition"):
            if(positionChannels[1] == "Xposition"):
                return "ZXY"
            if(positionChannels[1] == "Yposition"):
                return "ZYX"

    def getRotationChannelsOrder(self):
        if("rotation" not in self.channels[0] and len(self.channels) <= 3):
            print(f"\033[1;33mWARNING\033[0m: joint {self.name} has no rotation channels")
            return ""
        rotationChannels = self.channels[0:3] if("rotation" in self.channels[0] or "rotation" in self.channels[1] or "rotation" in self.channels[2]) else self.channels[3:6]
        if(rotationChannels[0] == "Xrotation"):
            if(rotationChannels[1] == "Yrotation"):
                return "XYZ"
            if(rotationChannels[1] == "Zrotation"):
                return "XZY"
        if(rotationChannels[0] == "Yrotation"):
            if(rotationChannels[1] == "Xrotation"):
                return "YXZ"
            if(rotationChannels[1] == "Zrotation"):
                return "YZX"
        if(rotationChannels[0] == "Zrotation"):
            if(rotationChannels[1] == "Xrotation"):
                return "ZXY"
            if(rotationChannels[1] == "Yrotation"):
                return "ZYX"
            
    def getChannelIndex(self, channelName):
        if(channelName not in self.channels):
            print(f"\033[1;33mWARNING\033[0m: joint {self.name} does not have channel {channelName}")
            return -1
        return self.channels.index(channelName)
    
    def getRotationFromOffset(self, canonicalRotation):
        offset = np.array(self.offset)
        offsetNormalized = offset / np.linalg.norm(offset)
        axis = np.cross(canonicalRotation, offsetNormalized)
        angle = np.arccos(np.clip(np.dot(canonicalRotation, offsetNormalized), -1.0, 1.0))

        if(np.linalg.norm(axis) < 1e-6):
            return R.identity()
        else:
            axis = axis / np.linalg.norm(axis)
            return R.from_rotvec(angle * axis)

class Skeleton:
    def __init__(self, rootJoint):
        self.root = rootJoint
        self.joints = self.buildJointDict(rootJoint)
        self.jointIndexes = self.buildJointIndexDict(rootJoint, [0])
        self.hierarchyIndexes = self.buildHierarchyIndexDict(rootJoint, [0])

    def buildJointDict(self, joint):
        jointDict = {joint.name: joint}
        for child in joint.children:
            jointDict.update(self.buildJointDict(child))
        return jointDict

    def buildJointIndexDict(self, joint, currentChannelIndex=None, jointIndex = None):
        if currentChannelIndex is None:
            currentChannelIndex = [0]
        if jointIndex is None:
            jointIndex = [0]

        jointIndexDict = {joint.name: currentChannelIndex[0]}
        joint.motionIndex = currentChannelIndex[0]
        joint.index = jointIndex[0]
        currentChannelIndex[0] += joint.getChannelCount()
        jointIndex[0] += 1
        for child in joint.children:
            jointIndexDict.update(self.buildJointIndexDict(child, currentChannelIndex, jointIndex))
        return jointIndexDict

    def buildHierarchyIndexDict(self, joint, currentChannelIndex=[0]):
        if(joint.parent != None):
            jointHierarchyIndexDict = {joint.name: joint.parent.index}
        else:
            jointHierarchyIndexDict = {joint.name: -1}
        for child in joint.children:
            jointHierarchyIndexDict.update(self.buildHierarchyIndexDict(child, currentChannelIndex))
        return jointHierarchyIndexDict

    def getJoint(self, jointName):
        return self.joints[jointName]

    def getJointIndex(self, jointName):
        return self.jointIndexes[jointName]

    def getJointIndexesList(self):
        return list(self.jointIndexes.values())
    
    def getJointHierarchyIndex(self, jointName):
        return self.hierarchyIndexes[jointName]

    def getHierarchyIndexesList(self):
        return list(self.hierarchyIndexes.values())

    def printJoint(self, node, prefix='', verbose = False):
        if node.parent == None:
            if not verbose:
                print(f"\033[1;32m{node.name} {node.index}\033[0m")
            else:
                print(f"\033[1;32m{node.name} {node.index}\033[0m: \033[1;34mChannels\033[0m: \033[36m{node.channels}\033[0m, \033[1;33mOffset\033[0m: \033[33m{node.offset}\033[0m")
        children = node.children
        for i, child in enumerate(children):
            is_last = (i == len(children) - 1)
            connector = '└── ' if is_last else '├── '
            child_prefix = prefix + ('    ' if is_last else '│   ')
            if not verbose:
                print(f"\033[1;32m{prefix + connector + child.name} {child.index}\033[0m")
            else:
                print(f"\033[1;32m{prefix + connector + child.name} {child.index}\033[0m: \033[1;34mChannels\033[0m: \033[36m{child.channels}\033[0m, \033[1;33mOffset\033[0m: \033[33m{child.offset}\033[0m")
            self.printJoint(child, child_prefix, verbose=verbose)

    def printSkeleton(self, verbose = False):
        self.printJoint(self.root, verbose=verbose)

class MotionData:
    def __init__(self, numFrames, frameTime, frames):
        if(numFrames != len(frames)):
            print("\033[1;33mWARNING\033[0m: Number of frames does not match number of frames in data. Taking the length of the motion data.")
        self.numFrames = len(frames)
        self.frameTime = frameTime
        self.frames = frames

    def addFrame(self, frameData):
        self.frames.append(frameData)

    def getFrame(self, frameIndex):
        return self.frames[frameIndex]
    
    def getFrameSlice(self, startFrame, endFrame):
        return self.frames[startFrame:endFrame]

    def getValues(self, valueIndex):
        return [x[valueIndex] for x in self.frames]
    
    def getValuesSlice(self, valueIndex, startFrame, endFrame):
        return [x[valueIndex] for x in self.frames[startFrame:endFrame]]

    def getValueAtFrame(self, valueIndex, frame):
        return self.frames[frame][valueIndex]
    
    def getValuesByJoint(self, joint):
        jointIndex = joint.motionIndex
        return [x[jointIndex:jointIndex + joint.getChannelCount()] for x in self.frames]

    def printHead(self, headSize = 10, verbose = False):
        print(f"\033[1;32mMOTION DATA\033[0m")
        print(f"\033[1;32mNumber of frames:\033[0m {self.numFrames}")
        print(f"\033[1;32mNumber of channels:\033[0m {len(self.frames[0])}")
        print(f"\033[1;32mFrame time:\033[0m {self.frameTime}")
        print(f"\033[1;32mMotion dataframe size:\033[0m {self.numFrames} x {len(self.frames[0])}")
        print(f"\033[1;32mHEAD\033[0m")
        for i in range(headSize):
            if not verbose:
                print(f"{self.frames[i][0:6]} ... {self.frames[i][-6:]}")
            else:
                print(f"{self.frames[i]}")
class BVHData:
    def __init__(self, skeleton, motion):
        self.skeleton = skeleton
        self.motion = motion
        self.skeletonDims = self.calculateSkeletonDims()
        self.motionDims = None
        
    def getJointLocalTransformAtFrame(self, jointName, frame, rotationMode = "Euler"):
        joint = self.skeleton.getJoint(jointName)
        jointIndex = self.skeleton.getJointIndex(jointName)
        r = None
        Xpos, Ypos, Zpos = 0.0, 0.0, 0.0
        if("Xrotation" in joint.channels and "Yrotation" in joint.channels and "Zrotation" in joint.channels):
            rotOrder = joint.getRotationChannelsOrder()
            angles = []
            for axis in rotOrder:
                axisName = axis + "rotation"
                if(axisName in joint.channels):
                    idx = jointIndex + joint.channels.index(axisName)
                    angles.append(self.motion.getValueAtFrame(idx, frame))
            r = R.from_euler(rotOrder, angles, degrees=True)
        if("Xposition" in joint.channels and "Yposition" in joint.channels and "Zposition" in joint.channels):
            Xpos = self.motion.getValueAtFrame(jointIndex + joint.channels.index("Xposition"), frame)
            Ypos = self.motion.getValueAtFrame(jointIndex + joint.channels.index("Yposition"), frame)
            Zpos = self.motion.getValueAtFrame(jointIndex + joint.channels.index("Zposition"), frame)

        if(r is None):
            if(rotationMode == "Euler"):
                return R.identity().as_euler('XYZ', degrees=True), [Xpos, Ypos, Zpos]
            if(rotationMode == "Quaternion"):
                return R.identity().as_quat(), [Xpos, Ypos, Zpos]
            if(rotationMode == "Matrix"):
                return R.identity().as_matrix(), [Xpos, Ypos, Zpos]
        else:
            if(rotationMode == "Euler"):
                return r.as_euler('XYZ', degrees=True), [Xpos, Ypos, Zpos]
            if(rotationMode == "Quaternion"):
                return r.as_quat(), [Xpos, Ypos, Zpos]
            if(rotationMode == "Matrix"):
                return r.as_matrix(), [Xpos, Ypos, Zpos]

    def calculateSkeletonDims(self):
        minX, minY, minZ = float('inf'), float('inf'), float('inf')
        maxX, maxY, maxZ = float('-inf'), float('-inf'), float('-inf')

        fkData0 = self.getFKAtFrame(0)
        for jointName, (rot, pos) in fkData0.items():
            # Extract the position of each joint
            x, y, z = pos
            
            # Update the min and max values for each axis (X, Y, Z)
            minX = min(minX, x)
            minY = min(minY, y)
            minZ = min(minZ, z)

            maxX = max(maxX, x)
            maxY = max(maxY, y)
            maxZ = max(maxZ, z)

        # Calculate height, width, and depth
        height = maxY - minY  # Difference in the Y-axis (vertical)
        width = maxX - minX   # Difference in the X-axis (horizontal)
        depth = maxZ - minZ   # Difference in the Z-axis (depth)

        return [height, width, depth]

    def getSkeletonDim(self, dimName):
        if(dimName == "width"):
            return self.skeletonDims[0]
        if(dimName == "height"):
            return self.skeletonDims[1]
        if(dimName == "depth"):
            return self.skeletonDims[2]

    def getSkeletonDims(self):
        return self.skeletonDims
    
    def calculateMotionDims(self):
        minX, minY, minZ = float('inf'), float('inf'), float('inf')
        maxX, maxY, maxZ = float('-inf'), float('-inf'), float('-inf')

        for frameIndex in range(self.motion.numFrames):
            fkDataRoot = self.getFKAtFrame(frameIndex)[self.skeleton.root.name][1]
            # Extract the position of each joint
            x, y, z = fkDataRoot
            
            # Update the min and max values for each axis (X, Y, Z)
            minX = min(minX, x)
            minY = min(minY, y)
            minZ = min(minZ, z)

            maxX = max(maxX, x)
            maxY = max(maxY, y)
            maxZ = max(maxZ, z)

        return [minX, maxX, minY, maxY, minZ, maxZ]

    def getMotionDims(self):
        if(self.motionDims is None):
            self.motionDims = self.calculateMotionDims()
        return self.motionDims
    
    def getChildFKAtFrame(self, joint, frame, parentTransform, fkFrame):
        localRot, localPos = self.getJointLocalTransformAtFrame(joint.name, frame, "Matrix")
        jointGlobalRot = np.matmul(parentTransform[0], localRot)
        rotatedOffset = np.matmul(parentTransform[0], joint.offset)
        if(any(ch in joint.channels for ch in ["Xposition", "Yposition", "Zposition"]) and joint == self.skeleton.root):
            jointGlobalPos = np.add(np.add(rotatedOffset, localPos), parentTransform[1])
        else:
            jointGlobalPos = np.add(rotatedOffset, parentTransform[1])
        fkFrame.update({joint.name: (jointGlobalRot, jointGlobalPos)})
        for child in joint.children:
            self.getChildFKAtFrame(child, frame, (jointGlobalRot, jointGlobalPos), fkFrame)

    def getFKAtFrame(self, frame):
        rootJoint = self.skeleton.root
        rootLocalRot, rootLocalPos = self.getJointLocalTransformAtFrame(rootJoint.name, frame, "Matrix")
        fkFrame = {rootJoint.name: (rootLocalRot, rootLocalPos)}
        for child in rootJoint.children:
            self.getChildFKAtFrame(child, frame, (rootLocalRot, rootLocalPos), fkFrame)
        return fkFrame
    
    def getFKAtFrameNormalized(self, frame, skeletonDim = "height"):
        fkFrame = self.getFKAtFrame(frame)
        normalizer = self.getSkeletonDim(skeletonDim)
        for jointName, (rot, pos) in fkFrame.items():
            fkFrame[jointName] = (rot, pos / normalizer)
        return fkFrame
    
    def writeJoint(self, joint, indent = 0):
        lines = []
        tab = '\t' * indent

        # An end site has to be written if and only if, the parent of the deleted joint has no children anymore
        # If we delete a joint, it's parent may still have children, so we don't need to write an end site
        if(joint.parent):
            if(len(joint.parent.children) == 0):
                lines.append(f"{tab}End Site")
                lines.append(f"{tab}{{")
                lines.append(f"\t{tab}OFFSET {' '.join(f'{x:.6f}' for x in joint.offset)}")
                lines.append(f"{tab}}}")
                return lines

        prefix = "ROOT" if indent == 0 else "JOINT"
        lines.append(f"{tab}{prefix} {joint.name}")
        lines.append(f"{tab}{{")
        lines.append(f"\t{tab}OFFSET {' '.join(f'{x:.6f}' for x in joint.offset)}")
        lines.append(f"\t{tab}CHANNELS {len(joint.channels)} {' '.join(map(str, joint.channels))}")

        if(len(joint.children) > 0):
            for child in joint.children:
                lines.extend(self.writeJoint(child, indent + 1))

        lines.append(f"{tab}}}")
        return lines

    def getHeader(self):
        header = ["HIERARCHY"]
        header.extend(self.writeJoint(self.skeleton.root, 0))
        header.append("MOTION")
        header.append(f"Frames: {self.motion.numFrames}")
        header.append(f"Frame Time: {self.motion.frameTime}")
        return header
    
    def rewriteHeaderOffsets(self):
        jointName = ""
        for lineIndex, line in enumerate(self.header):
            if("ROOT" in line or "JOINT" in line): 
                jointName = line.split(" ")[-1]
            if("End Site" in line):
                jointName = jointName + "_EndSite"
            
            if("OFFSET" in line):
                newValuesFormatted = ['{:+0.6f}'.format(v) if v < 0 else '{:0.6f}'.format(v) for v in self.skeleton.getJoint(jointName).offset]
                line = re.sub(r'([-+]?\d*\.\d{6})\s+([-+]?\d*\.\d{6})\s+([-+]?\d*\.\d{6})$', ' '.join(newValuesFormatted), line)
                self.header[lineIndex] = line

    def getRestPoseJoint(self, joint, canonicalRotation, poseDict):
        poseDict.update({joint.name: joint.getRotationFromOffset(canonicalRotation)})
        for child in joint.children:
            if("EndSite" not in child.name):
                self.getRestPoseJoint(child, canonicalRotation, poseDict)

    def getRestPose(self, canonicalAxis = "Y"):
        if(canonicalAxis == "X"):
            canonicalRotation = np.array([1, 0, 0])
        elif(canonicalAxis == "Y"):
            canonicalRotation = np.array([0, 1, 0])
        elif(canonicalAxis == "Z"):
            canonicalRotation = np.array([0, 0, 1])
        else:
            print("ERROR: Invalid canonical axis. The canonical axis has to be either X, Y or Z. Default: Y.")
        root = self.skeleton.root
        poseDict = dict()
        self.getRestPoseJoint(root, canonicalRotation, poseDict)
        return poseDict
    
    def applyOffsetToChildren(self, joint, rNew):
        for child in joint.children:
            length = np.linalg.norm(child.offset)
            canonical = np.array([0, 1, 0])
            child.offset = rNew.apply(canonical * length)
            self.applyOffsetToChildren(child, rNew)

    def applyRotationToItselfAndChildren(self, joint, oldPose, newPose, rNew):
        for frame in self.motion.frames:
            rotationChannels = [0, 1, 2] if("rotation" in joint.channels[0] or "rotation" in joint.channels[1] or "rotation" in joint.channels[2]) else [3, 4, 5]
            oldRotation = R.from_euler(joint.getRotationChannelsOrder(), [frame[self.skeleton.getJointIndex(joint.name) + rotationChannels[0]],
                                                                    frame[self.skeleton.getJointIndex(joint.name) + rotationChannels[1]],
                                                                    frame[self.skeleton.getJointIndex(joint.name) + rotationChannels[2]]], degrees=True)
            rOld = oldPose[joint.name]
            newRotation = (rNew * rOld.inv() * oldRotation).as_euler(joint.getRotationChannelsOrder(), degrees = True)
            
            print(f"Joint: {joint.name}")
            print(f"Old Pose: {rOld.as_matrix()}")
            print(f"New Pose: {rNew.as_matrix()}")
            print(f"Old Rotation: {oldRotation.as_euler(joint.getRotationChannelsOrder(), degrees=True)}")
            print(f"New Rotation: {newRotation}")
        
            frame[self.skeleton.getJointIndex(joint.name) + rotationChannels[0]] = newRotation[0]
            frame[self.skeleton.getJointIndex(joint.name) + rotationChannels[1]] = newRotation[1]
            frame[self.skeleton.getJointIndex(joint.name) + rotationChannels[2]] = newRotation[2]
        # for child in joint.children:
        #     if(not "_EndSite" in child.name):
        #         rNew = newPose[child.name]
        #         self.applyRotationToItselfAndChildren(child, oldPose, newPose, rNew)

    def setRestPose(self, poseDict):
        oldPose = copy.deepcopy(self.getRestPose())
        newPose = {}
        for poseName, pose in poseDict.items():
            newPose[poseName] = R.from_euler('XYZ', poseDict[poseName], degrees=True)

        for joint in self.skeleton.joints.values():
            if(joint.name in poseDict.keys()):
                rNew = newPose[joint.name]
                self.applyOffsetToChildren(joint, rNew)
                # newPose = self.getRestPose()
                self.applyRotationToItselfAndChildren(joint, oldPose, newPose, rNew)

        self.rewriteHeaderOffsets()


from bvhIO import readBvh, writeBvh, writeBvhToCsv, writePositionsToCsv
from bvhManipulation import centerSkeletonRoot, centerSkeletonFeet, centerSkeletonAroundJoint, rotateSkeletonWorld, rotateSkeletonLocal, centerSkeletonXZ, moveSkeleton
from bvhSlicer import getBvhSlices, groupBvhSlices, appendBvhSlices
from bvhVisualizerMpl import showBvhAnimation
from bvhSkeletonEditor import scaleSkeleton
if __name__ == "__main__":
    bvhData = readBvh("testBvhFiles/test.bvh")
    bvhDataNew = scaleSkeleton(bvhData, 1.1)
    writeBvh(bvhDataNew, "testBvhFiles/testScaled.bvh")
    # bvhData.setRestPose({'LeftLeg': [ 1.31714436e-05, 90, -9.00000553e+01]})
    # writeBvh(bvhData, "testBvhFiles/testNewRest.bvh")


import copy
from scipy.spatial.transform import Rotation as R

def centerSkeletonRoot(bvhData, fkFrame=0):
    bvhDataCopy = copy.deepcopy(bvhData)
    frame = bvhDataCopy.motion.getFrame(fkFrame)
    rootIndex = bvhDataCopy.skeleton.getJointIndex(bvhDataCopy.skeleton.root.name)
    offsets = [-float(frame[rootIndex]), -float(frame[rootIndex + 1]), -float(frame[rootIndex + 2])]
    for frame in bvhDataCopy.motion.frames:
        frame[rootIndex] += offsets[0]
        frame[rootIndex+1] += offsets[1]
        frame[rootIndex+2] += offsets[2]

    return bvhDataCopy

def centerSkeletonFeet(bvhData, leftFootName = "LeftFoot", rightFootName = "RightFoot", fkFrame=0):
    bvhDataCopy = copy.deepcopy(bvhData)
    if(leftFootName not in bvhDataCopy.skeleton.joints):
        raise Exception(f"Left foot name ({leftFootName}) not found in skeleton")
    if(rightFootName not in bvhDataCopy.skeleton.joints):
        raise Exception(f"Right foot name ({rightFootName}) not found in skeleton")
    avgFootHeight = (bvhDataCopy.getFKAtFrame(fkFrame)[leftFootName][1][1] + bvhDataCopy.getFKAtFrame(fkFrame)[rightFootName][1][1]) / 2
    avgRootHeight = bvhDataCopy.getFKAtFrame(fkFrame)[bvhDataCopy.skeleton.root.name][1][1]
    frame = bvhDataCopy.motion.getFrame(fkFrame)
    rootIndex = bvhDataCopy.skeleton.getJointIndex(bvhDataCopy.skeleton.root.name)
    offsets = [-float(frame[rootIndex]), -float(frame[rootIndex + 1]) + (avgRootHeight - avgFootHeight), -float(frame[rootIndex + 2])]
    for frame in bvhDataCopy.motion.frames:
        frame[rootIndex] += offsets[0]
        frame[rootIndex+1] += offsets[1]
        frame[rootIndex+2] += offsets[2]

    return bvhDataCopy

def centerSkeletonXZ(bvhData, fkFrame=0):
    bvhDataCopy = copy.deepcopy(bvhData)
    frame = bvhDataCopy.motion.getFrame(fkFrame)
    rootIndex = bvhDataCopy.skeleton.getJointIndex(bvhDataCopy.skeleton.root.name)
    offsets = [-float(frame[rootIndex]), -float(frame[rootIndex + 1]), -float(frame[rootIndex + 2])]
    for frame in bvhDataCopy.motion.frames:
        frame[rootIndex] += offsets[0]
        frame[rootIndex+2] += offsets[2]

    return bvhDataCopy

def centerSkeletonAroundJoint(bvhData, jointName, fkFrame=0):
    bvhDataCopy = copy.deepcopy(bvhData)
    if(jointName not in bvhDataCopy.skeleton.joints):
        raise Exception(f"Selected joint ({jointName}) not found in skeleton")
    
    forwardFrame = bvhDataCopy.getFKAtFrame(fkFrame)
    frame = bvhDataCopy.motion.getFrame(fkFrame)
    jointOffsets = forwardFrame[jointName][1] - forwardFrame[bvhDataCopy.skeleton.root.name][1]
    rootIndex = bvhDataCopy.skeleton.getJointIndex(bvhDataCopy.skeleton.root.name)
    offsets = [-float(frame[rootIndex]) - jointOffsets[0], -float(frame[rootIndex + 1]) - jointOffsets[1], -float(frame[rootIndex + 2]) - jointOffsets[2]]
    for frame in bvhDataCopy.motion.frames:
        frame[rootIndex] += offsets[0]
        frame[rootIndex+1] += offsets[1]
        frame[rootIndex+2] += offsets[2]

    return bvhDataCopy

def rotateSkeletonLocal(bvhData, angle, fkFrame=0):
    if(len(angle) != 3):
        raise Exception("angle must be a list of length 3")
    bvhDataCopy = copy.deepcopy(bvhData)
    rotation = R.from_euler('XYZ', [angle[0], angle[1], angle[2]], degrees=True)
    rootIndex = bvhDataCopy.skeleton.getJointIndex(bvhDataCopy.skeleton.root.name)
    originPoint = bvhDataCopy.getFKAtFrame(fkFrame)[bvhDataCopy.skeleton.root.name][1]
    rootChannelOrder = bvhData.skeleton.root.getRotationChannelsOrder()
    for frameIndex, frame in enumerate(bvhDataCopy.motion.frames):
        fkFrameRootPos = bvhDataCopy.getFKAtFrame(frameIndex)[bvhDataCopy.skeleton.root.name][1]
        newPos = [x - y for x,y in zip(fkFrameRootPos, originPoint)]
        newPos = rotation.apply(newPos)
        newPos = [x + y for x,y in zip(newPos, originPoint)]
        baseRotation = R.from_euler(rootChannelOrder, frame[rootIndex+3:rootIndex+6], degrees=True)
        newRotation = rotation * baseRotation
        frame[rootIndex:rootIndex+3] = newPos
        frame[rootIndex+3:rootIndex+6] = newRotation.as_euler(rootChannelOrder, degrees=True)
    return bvhDataCopy

def rotateSkeletonWorld(bvhData, angle):
    if(len(angle) != 3):
        raise Exception("angle must be a list of length 3")
    bvhDataCopy = copy.deepcopy(bvhData)
    rotation = R.from_euler('XYZ', [angle[0], angle[1], angle[2]], degrees=True)
    rootIndex = bvhDataCopy.skeleton.getJointIndex(bvhDataCopy.skeleton.root.name)
    rootChannelOrder = bvhData.skeleton.root.getRotationChannelsOrder()
    for frameIndex, frame in enumerate(bvhDataCopy.motion.frames):
        fkFrameRootPos = bvhDataCopy.getFKAtFrame(frameIndex)[bvhDataCopy.skeleton.root.name][1]
        newPos = rotation.apply(fkFrameRootPos)
        baseRotation = R.from_euler(rootChannelOrder, frame[rootIndex+3:rootIndex+6], degrees=True)
        newRotation = rotation * baseRotation
        frame[rootIndex:rootIndex+3] = newPos
        frame[rootIndex+3:rootIndex+6] = newRotation.as_euler(rootChannelOrder, degrees=True)
    return bvhDataCopy

def moveSkeleton(bvhData, offsets):
    if(len(offsets) != 3):
        raise Exception("offsets must be a list of length 3")
    bvhDataCopy = copy.deepcopy(bvhData)
    rootIndex = bvhDataCopy.skeleton.getJointIndex(bvhDataCopy.skeleton.root.name)
    for frame in bvhDataCopy.motion.frames:
        frame[rootIndex] += offsets[0]
        frame[rootIndex+1] += offsets[1]
        frame[rootIndex+2] += offsets[2]

    return bvhDataCopy


from bvhTools.bvhDataTypes import Joint, Skeleton, MotionData, BVHData

def checkJointForPosition(joint, rootJoint, nonRootJointsWithPos):
    if joint != rootJoint and any(ch in joint.channels for ch in ["Xposition", "Yposition", "Zposition"]):
        nonRootJointsWithPos.append(joint.name)
    for child in joint.children:
        checkJointForPosition(child, rootJoint, nonRootJointsWithPos)

def buildBvhStructure(header, motion, numFrames, frameTime):
    currentIndex = 0
    jointIndex = 0
    rootJoint = None
    while(currentIndex < len(header)):
        if("ROOT" in header[currentIndex]):
            rootJoint, newIndex, jointIndex = readJoint(header, currentIndex, jointIndex)
            currentIndex = newIndex
            break
        currentIndex += 1
    skeleton = Skeleton(rootJoint)
    # Check if any the BVH joints have position channels, and throw a warning if so
    nonRootJointsWithPos = []
    checkJointForPosition(skeleton.root, rootJoint, nonRootJointsWithPos)
    if len(nonRootJointsWithPos) > 0:
        print(f"\033[1;33mWARNING\033[0m: The following joints have position channels: {', '.join(nonRootJointsWithPos)}. \nTheir positions will be ignored when calculating FK.\n")

    motionData = MotionData(numFrames=numFrames, frameTime=frameTime, frames = motion)
    bvh = BVHData(skeleton=skeleton, motion=motionData)
    return bvh

def readEndSite(header, currentIndex, jointIndex, parent):
    currentIndex += 1
    while(currentIndex < len(header)):
        if("{" in header[currentIndex]):
            currentIndex += 1
        if("OFFSET" in header[currentIndex]):
            offset = [float(x) for x in header[currentIndex].split(" ")[1:]]
            currentIndex += 1
        if("}" in header[currentIndex]):
            currentIndex += 1
            break
    
    endSite = Joint(name = f"{parent.name}_EndSite", index=jointIndex, offset=offset, channels = [], parent=parent)
    jointIndex += 1
    return endSite, currentIndex, jointIndex
        
def readJoint(header, currentIndex, jointIndex, parent=None):
    jointName = header[currentIndex].split(" ")[1]
    currentIndex += 1
    jointObject = Joint(name = jointName, index=jointIndex, offset=None, channels=[], parent = parent)
    jointIndex += 1
    while(currentIndex < len(header)):
        if("{" in header[currentIndex]):
            currentIndex += 1
        
        if("OFFSET" in header[currentIndex]):
            jointObject.setOffset([float(x) for x in header[currentIndex].rstrip().split(" ")[1:]])
            currentIndex += 1
        
        if("CHANNELS" in header[currentIndex]):
            jointObject.setChannels([str(x) for x in header[currentIndex].rstrip().split(" ")[2:]])
            currentIndex += 1
        
        if("JOINT" in header[currentIndex]):
            childJoint, currentIndex, jointIndex = readJoint(header, currentIndex, jointIndex, jointObject)
            jointObject.addChild(childJoint)

        if("End Site" in header[currentIndex]):
            endSite, currentIndex, jointIndex = readEndSite(header, currentIndex, jointIndex, jointObject)
            jointObject.addChild(endSite)
        
        if("}" in header[currentIndex]):
            currentIndex += 1
            break

    return jointObject, currentIndex, jointIndex

def readBvh(bvhPath):
    header = []
    motion = []
    numFrames = 0
    frameTime = 0.0

    with open(bvhPath, "r") as f:
        # read and process the header
        line = f.readline()
        while(True):
            if("MOTION" in line):
                break
            header.append(line.strip("\n"))
            line = f.readline()

        # read and process the motion data
        line = f.readline()
        while(line != ""):
            if("Frames:" in line):
                numFrames = int(line.split(" ")[1])
            elif("Frame Time:" in line):
                frameTime = float(line.split(" ")[2])
            else:
                motion.append([float(x) for x in line.rstrip().replace("\n", "").split(" ")])
            line = f.readline()
    bvhData = buildBvhStructure(header, motion, numFrames, frameTime)
    return bvhData

def writeBvh(bvhData, bvhPath, decimals = 6):
    with open(bvhPath, "w") as f:
        for line in bvhData.getHeader():
            f.write(line)
            f.write("\n")
        for frame in bvhData.motion.frames:
            strings = [f"{x:.6f}" for x in frame]
            for string in strings:
                f.write(string + " ")
            f.write("\n")

def writeBvhToCsv(bvhData, csvPath, decimals = 6):
    with open(csvPath, "w") as f:
        for joint in bvhData.skeleton.joints:
            jointObject = bvhData.skeleton.getJoint(joint)
            jointClasses = [jointObject.name +  "_" + str(channel) for channel in jointObject.channels]
            if(len(jointClasses) > 0):
                f.write(",".join(jointClasses) + ",")
        f.write("\n")
        for frame in bvhData.motion.frames:
            f.write(",".join([f"{x:.{decimals}f}" for x in frame]) + "\n")

def writePositionsToCsv(bvhData, csvPath, decimals = 6):
    with open(csvPath, "w") as f:
        fkFrame = bvhData.getFKAtFrame(0)
        f.write(",".join([str(x)+ "_x," + str(x)+"_y,"+ str(x)+"_z" for x in fkFrame.keys()]) + "\n")
        for frameIndex in range(bvhData.motion.numFrames):
            fkFrame = bvhData.getFKAtFrame(frameIndex)
            points = [x[1] for x in fkFrame.values()]
            f.write(",".join([f"{x[0]:.{decimals}f}, {x[1]:.{decimals}f}, {x[2]:.{decimals}f}" for x in points]) + "\n")


import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.widgets import Button, TextBox
from matplotlib import get_backend
import numpy as np

def showBvhAnimation(bvhData, showPoints = True, showLines = True, showQuivers = True, 
                     showLabels = False, pointColor = "#4287f5", pointMarker = "o", lineColor = "#666666", lineWidth = 2):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    manager = plt.get_current_fig_manager()
    backend = get_backend()

    # Try to maximize the window, but fail silently if it doesn't work
    try:
        if backend == 'TkAgg':
            try:
                manager.window.state('zoomed')  # Windows
            except Exception:
                manager.window.attributes('-zoomed', True)  # Linux (X11)
        elif backend == 'QtAgg':
            manager.window.showMaximized()
        elif backend == 'WXAgg':
            manager.frame.Maximize(True)
    except Exception as e:
        print(f"[Info] Could not maximize window: {e}")

    motionDims = bvhData.getMotionDims()
    maxDim = np.max(np.abs(motionDims))
    quiverSize = 0.05 * maxDim
    numFrames = bvhData.motion.numFrames
    frameTime = bvhData.motion.frameTime
    isPaused = [False]
    currentFrame = [0]
    ax.set_xlim3d(-maxDim, maxDim)
    ax.set_ylim3d(-maxDim, maxDim)
    ax.set_zlim3d(-maxDim, maxDim)

    parentList = bvhData.skeleton.getHierarchyIndexesList()
    labelList = list(bvhData.skeleton.joints.keys())

    def update(_):            
        for coll in ax.collections[:]:
            coll.remove()
        for line in ax.lines[:]:
            line.remove()
        for text in ax.texts[:]:
            text.remove()
        
        fkFrame = bvhData.getFKAtFrame(currentFrame[0])
        points = [x[1] for x in fkFrame.values()]
        if(showQuivers):
            ax.quiver(0, 0, 0, quiverSize, 0, 0, color='r', label='X')  # Red = X
            ax.quiver(0, 0, 0, 0, quiverSize, 0, color='g', label='Y')  # Green = Y
            ax.quiver(0, 0, 0, 0, 0, quiverSize, color='b', label='Z')  # Blue = Z

        if(showPoints):
            ax.scatter([-p[0] for p in points], [p[2] for p in points], [p[1] for p in points], c=pointColor, marker=pointMarker)
        
        if(showLabels):
            for index, point in enumerate(points):
                ax.text(-point[0], point[2], point[1], labelList[index])

        if(showLines):
            for index, parent in enumerate(parentList):
                if(parent!=-1):
                    ax.plot([-points[index][0], -points[parent][0]],
                            [points[index][2], points[parent][2]],
                            [points[index][1], points[parent][1]], color=lineColor, linewidth=lineWidth)

        if(not isPaused[0]):
            currentFrame[0] = (currentFrame[0] + 1) % numFrames
            label.set_text(f"Frame: {currentFrame[0]}")

    def togglePause(event):
        isPaused[0] = not isPaused[0]
        btnPlayPause.label.set_text("Play" if isPaused[0] else "Pause")
        textbox.set_val(str(currentFrame[0]))
        label.set_text(f"Frame: {currentFrame[0]}")

    def frameBack(event):
        isPaused[0] = True
        btnPlayPause.label.set_text("Play" if isPaused[0] else "Pause")
        currentFrame[0] -= 1
        if(currentFrame[0] < 0):
            currentFrame[0] = numFrames - 1
        textbox.set_val(str(currentFrame[0]))
        label.set_text(f"Frame: {currentFrame[0]}")

    def frameForward(event):
        isPaused[0] = True
        btnPlayPause.label.set_text("Play" if isPaused[0] else "Pause")
        currentFrame[0] += 1
        if(currentFrame[0] >= numFrames):
            currentFrame[0] = 0
        textbox.set_val(str(currentFrame[0]))
        label.set_text(f"Frame: {currentFrame[0]}")

    def faster(event):
        global anim
        newInterval = anim.event_source.interval * 0.5
        anim.event_source.stop()
        anim = animation.FuncAnimation(fig, update, frames=numFrames, interval=newInterval, repeat=True)
        plt.draw()

    def slower(event):
        global anim
        newInterval = anim.event_source.interval * 2
        anim.event_source.stop()
        anim = animation.FuncAnimation(fig, update, frames=numFrames, interval=newInterval, repeat=True)
        plt.draw()

    def goToFrame(text):
        if(int(text) < numFrames):
            currentFrame[0] = int(text)
        else:
            currentFrame[0] = numFrames - 1
        label.set_text(f"Frame: {currentFrame[0]}")

    global anim
    anim = animation.FuncAnimation(fig, update, frames=numFrames, interval=frameTime * 1000, repeat = True)

    axBtnPlayPause = plt.axes([0.45, 0.05, 0.1, 0.05])
    btnPlayPause = Button(axBtnPlayPause, "Pause")
    btnPlayPause.on_clicked(togglePause)

    axBtnBack = plt.axes([0.34, 0.05, 0.1, 0.05])
    btnBack = Button(axBtnBack, "Back")
    btnBack.on_clicked(frameBack)

    axBtnForward = plt.axes([0.56, 0.05, 0.1, 0.05])
    btnForward = Button(axBtnForward, "Forward")
    btnForward.on_clicked(frameForward)

    axBtnFaster = plt.axes([0.395, 0.9, 0.1, 0.05])
    btnFaster = Button(axBtnFaster, "Faster")
    btnFaster.on_clicked(faster)

    axBtnSlower = plt.axes([0.505, 0.9, 0.1, 0.05])
    btnSlower = Button(axBtnSlower, "Slower")
    btnSlower.on_clicked(slower)
    
    ax_textbox = plt.axes([0.8, 0.9, 0.1, 0.05])  # [x, y, width, height]
    textbox = TextBox(ax_textbox, "Go to frame: ")
    textbox.on_submit(goToFrame)

    label = fig.text(0.475, 0.85, "Frame: 0", fontsize=12)

    plt.show()


import copy
from bvhTools.bvhDataTypes import BVHData, MotionData

def getBvhSlice(bvhData, fromFrame, toFrame):
    if(fromFrame > toFrame):
        print(f"\033[1;33mWARNING\033[0m: fromFrame ({fromFrame}) must be less than toFrame ({toFrame}). Returning original bvh.")
        return bvhData
    if(fromFrame < 0 or fromFrame > bvhData.motion.numFrames):
        print(f"\033[1;33mWARNING\033[0m: fromFrame ({fromFrame}) is less than zero or out of range. Returning original bvh.")
        return bvhData
    if (toFrame < 0 or toFrame > bvhData.motion.numFrames):
        print(f"\033[1;33mWARNING\033[0m: toFrame ({toFrame}) is less than zero or out of range. Returning original bvh.")
    slicedBvh = BVHData(bvhData.skeleton, MotionData(toFrame - fromFrame, bvhData.motion.frameTime, bvhData.motion.getFrameSlice(fromFrame, toFrame)))
    return slicedBvh

def getBvhSlices(bvhData, fromFrames, toFrames):
    if(len(fromFrames) != len(toFrames)):
        print(f"\033[1;33mWARNING\033[0m: fromFrames ({len(fromFrames)}) and toFrames ({len(toFrames)}) must be the same length. Returning original bvh.")
        return bvhData
    if(any(f < 0 for f in fromFrames) or any(f > bvhData.motion.numFrames for f in fromFrames)):
        oor = [f for f in fromFrames if f < 0 or f > bvhData.motion.numFrames]
        print(f"\033[1;33mWARNING\033[0m: some fromFrame ({oor}) is out of range. Returning original bvh.")
        return bvhData
    if(any(f < 0 for f in toFrames) or any(f > bvhData.motion.numFrames for f in toFrames)):
        oor = [f for f in toFrames if f < 0 or f > bvhData.motion.numFrames]
        print(f"\033[1;33mWARNING\033[0m: some toFrame ({oor}) is out of range. Returning original bvh.")
        return bvhData
    bvhsToReturn = []
    for fromFrame, toFrame in zip(fromFrames, toFrames):
        bvhsToReturn.append(getBvhSlice(bvhData, fromFrame, toFrame))
    return bvhsToReturn

def appendBvhSlices(baseBvh, bvhsToAppend):
    if(len(bvhsToAppend) == 0):
        print(f"\033[1;33mWARNING\033[0m: You must provide at least one BVH to append. Returning original bvh.")
        return baseBvh
    bvhData = copy.deepcopy(baseBvh)
    for bvh in bvhsToAppend:
        for frame in bvh.motion.frames:
            bvhData.motion.frames.append(frame)
        bvhData.motion.numFrames += bvh.motion.numFrames
    return bvhData
        
def groupBvhSlices(bvhsToGroup):
    if(len(bvhsToGroup) <= 1):
        print(f"\033[1;33mWARNING\033[0m: You must provide at least two BVHs to append. Returning original bvh.")
        return bvhsToGroup[0]
    bvhData = copy.deepcopy(bvhsToGroup[0])
    for bvh in bvhsToGroup[1:]:
        for frame in bvh.motion.frames:
            bvhData.motion.frames.append(frame)
        bvhData.motion.numFrames += bvh.motion.numFrames
    return bvhData